import sys
import base64
import hashlib
import urllib
import json
import webbrowser
import secrets
import threading
import logging
from time import sleep
import requests

from flask import Flask, request
from werkzeug.serving import make_server

from suite_py.lib import logger

## Global vars for comunication between flask thread and main cli function
received_callback = None
code = None
error_message = None
received_state = None

## Global Flask App
app = Flask(__name__)
log = logging.getLogger("werkzeug")
log.disabled = True


@app.route("/callback")
def callback():
    """
    The callback is invoked after a completed login attempt (succesful or otherwise).
    It sets global variables with the auth code or error messages, then sets the
    polling flag received_callback.
    :return:
    """
    global received_callback, code, error_message, received_state
    error_message = None
    code = None
    if "error" in request.args:
        error_message = request.args["error"] + ": " + request.args["error_description"]
    else:
        code = request.args["code"]
    received_state = request.args["state"]
    received_callback = True
    return "Please return to your application now."


class ServerThread(threading.Thread):
    """
    The Flask server is done this way to allow shutting down after a single request has been received.
    """

    def __init__(self, app):
        threading.Thread.__init__(self)
        self.srv = make_server("127.0.0.1", 5000, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        # logger.debug('starting server')
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()


class Login:
    def __init__(self, config):

        try:
            if (
                not config.auth0["client_id"]
                or not config.auth0["tenant"]
                or not config.auth0["scope"]
            ):
                self.usage()
                sys.exit(-1)
        except AttributeError:
            self.usage()
            sys.exit(-1)

        self._config = config
        self.auth0_client_id = config.auth0["client_id"]
        self.auth0_tenant = config.auth0["tenant"]
        self.auth0_scope = config.auth0["scope"]
        self.redirect_uri = "http://127.0.0.1:5000/callback"

    def usage(self):
        logger.warning("Unable to login: missing config")
        logger.warning(
            "please check docs: https://github.com/primait/suite_py/blob/master/README.md"
        )

    def auth0_url_encode(self, byte_data):
        """
        Safe encoding handles + and /, and also replace = with nothing
        :param byte_data:
        :return:
        """
        return base64.urlsafe_b64encode(byte_data).decode("utf-8").replace("=", "")

    def generate_challenge(self, a_verifier):
        return self.auth0_url_encode(hashlib.sha256(a_verifier.encode()).digest())

    def run(self):
        global received_callback, code, error_message, received_state

        # from https://auth0.com/docs/flows/add-login-using-the-authorization-code-flow-with-pkce
        # Step1: Create code verifier: Generate a code_verifier that will be sent to Auth0 to request tokens.
        verifier = self.auth0_url_encode(secrets.token_bytes(32))
        ## Step2: Create code challenge: Generate a code_challenge from the code_verifier that will be sent to Auth0 to request an authorization_code.
        challenge = self.generate_challenge(verifier)
        state = self.auth0_url_encode(secrets.token_bytes(32))
        client_id = self.auth0_client_id
        tenant = self.auth0_tenant
        scope = self.auth0_scope
        redirect_uri = self.redirect_uri
        #
        # We generate a nonce (state) that is used to protect against attackers invoking the callback
        base_url = "https://%s.auth0.com/authorize?" % tenant
        url_parameters = {
            "audience": "qainit",
            "scope": scope,
            "response_type": "code",
            "redirect_uri": redirect_uri,
            "client_id": client_id,
            "code_challenge": challenge.replace("=", ""),
            "code_challenge_method": "S256",
            "state": state,
        }
        url = base_url + urllib.parse.urlencode(url_parameters)

        # Step3: Authorize user: Request the user's authorization and redirect back to your app with an authorization_code.
        # Open the browser window to the login url
        # Start the server
        # Poll til the callback has been invoked
        received_callback = False
        webbrowser.open_new(url)
        server = ServerThread(app)
        server.start()
        while not received_callback:
            sleep(1)
        server.shutdown()

        if state != received_state:
            print(
                "Error: session replay or similar attack in progress. Please log out of all connections."
            )
            sys.exit(1)

        if error_message:
            print("An error occurred:")
            print(error_message)
            sys.exit(1)

        # Step4: Request tokens: Exchange your authorization_code and code_verifier for tokens.
        url = "https://%s.auth0.com/oauth/token" % tenant
        headers = {"Content-Type": "application/json"}
        body = {
            "grant_type": "authorization_code",
            "client_id": client_id,
            "code_verifier": verifier,
            "code": code,
            "audience": "https://prima-it-staging.eu.auth0.com/api/v2/",
            "redirect_uri": redirect_uri,
        }
        r = requests.post(url, headers=headers, data=json.dumps(body))
        data = r.json()
        if "access_token" in data:
            token = str(data["access_token"])
        else:
            logger.error("access_token not found in auth0 response")
            sys.exit(1)
        logger.info("login succeded")
        self._config.put_cache("auth0_token", token)



# -*- encoding: utf-8 -*-
import os
import subprocess
from subprocess import CalledProcessError
import sys

from suite_py.lib import logger


class VaultHandler:
    def __init__(self, repo, config):
        self._repo = repo
        self._projects_home = config.user["projects_home"]
        self._path = os.path.join(self._projects_home, repo)

    def exec(self, profile, command, additional_args=""):
        try:
            c = subprocess.Popen(
                f"aws-vault exec {profile} {additional_args} -- {command}",
                stdout=subprocess.PIPE,
                shell=True,
                cwd=self._path,
            )  # .stdout.read()
            c.wait()
            return c
        except (CalledProcessError) as e:
            logger.error(f"Error during command execution: {e}")
            sys.exit(-1)



#! -*- encoding: utf-8 -*-
import os
import sys

import click
from autoupgrade import Package

from suite_py.__version__ import __version__
from suite_py.commands.ask_review import AskReview
from suite_py.commands.check import Check
from suite_py.commands.create_branch import CreateBranch
from suite_py.commands.create_qa import CreateQA
from suite_py.commands.delete_qa import DeleteQA
from suite_py.commands.release import Release
from suite_py.commands.deploy import Deploy
from suite_py.commands.generator import Generator
from suite_py.commands.id import ID
from suite_py.commands.ip import IP
from suite_py.commands.merge_pr import MergePR
from suite_py.commands.open_pr import OpenPR
from suite_py.commands.project_lock import ProjectLock
from suite_py.commands.status import Status
from suite_py.commands.aggregator import Aggregator
from suite_py.commands.secret import Secret
from suite_py.commands.login import Login
from suite_py.lib.config import Config
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.tokens import Tokens


@click.group()
@click.option(
    "--project",
    type=click.Path(exists=True),
    default=os.getcwd(),
    help="Path of the project to run the command on (the default is current directory)",
)
@click.option(
    "--timeout",
    type=click.INT,
    help="Timeout in seconds for Captainhook operations",
)
@click.pass_context
def main(ctx, project, timeout):
    Package("suite_py").upgrade()
    print(f"v{__version__}")

    config = Config()

    if not git.is_repo(project):
        print(f"the folder {project} is not a git repo")
        sys.exit(-1)

    if not os.path.basename(project) in os.listdir(config.user["projects_home"]):
        print(f"the folder {project} is not in {config.user['projects_home']}")
        sys.exit(-1)

    if not config.user.get("skip_confirmation", False) and not prompt_utils.ask_confirm(
        f"Do you want to continue on project {os.path.basename(project)}?"
    ):
        sys.exit()

    ctx.ensure_object(dict)
    ctx.obj["project"] = os.path.basename(project)
    if timeout:
        config.user["captainhook_timeout"] = timeout
    ctx.obj["config"] = config
    ctx.obj["tokens"] = Tokens()
    os.chdir(os.path.join(config.user["projects_home"], ctx.obj["project"]))


@main.command(
    "create-branch", help="Create local branch and set the YouTrack card in progress"
)
@click.option("--card", type=click.STRING, help="YouTrack card number (ex. PRIMA-123)")
@click.pass_obj
def cli_create_branch(obj, card):
    CreateBranch(obj["project"], card, obj["config"], obj["tokens"]).run()


@main.command("lock", help="Lock project on staging or prod")
@click.argument(
    "environment", type=click.Choice(("staging", "production", "deploy", "merge"))
)
@click.pass_obj
def cli_lock_project(obj, environment):
    ProjectLock(obj["project"], environment, "lock", obj["config"]).run()


@main.command("unlock", help="Unlock project on staging or prod")
@click.argument(
    "environment", type=click.Choice(("staging", "production", "deploy", "merge"))
)
@click.pass_obj
def cli_unlock_project(obj, environment):
    ProjectLock(obj["project"], environment, "unlock", obj["config"]).run()


@main.command("open-pr", help="Open a PR on GitHub")
@click.pass_obj
def cli_open_pr(obj):
    OpenPR(obj["project"], obj["config"], obj["tokens"]).run()


@main.command("ask-review", help="Requests a PR review")
@click.pass_obj
def cli_ask_review(obj):
    AskReview(obj["project"], obj["config"], obj["tokens"]).run()


@main.command("create-qa", help="Create QA env (integration with qainit)")
@click.pass_obj
def cli_create_qa(obj):
    CreateQA(obj["project"], obj["config"], obj["tokens"]).run()


@main.command("delete-qa", help="Delete QA env (integration with qainit)")
@click.pass_obj
def cli_delete_qa(obj):
    DeleteQA(obj["project"], obj["config"], obj["tokens"]).run()


@main.command(
    "merge-pr", help="Merge the selected branch to master if all checks are OK"
)
@click.pass_obj
def cli_merge_pr(obj):
    MergePR(obj["project"], obj["config"], obj["tokens"]).run()


@main.group("release", help="Manage releases")
def release():
    pass


@release.command("create", help="Create a github release")
@click.option("--deploy", is_flag=True, help="Trigger deploy after release creation")
@click.pass_obj
def cli_release_create(obj, deploy):
    Release(
        "create", obj["project"], obj["config"], obj["tokens"], flags={"deploy": deploy}
    ).run()


@release.command("deploy", help="Deploy a github release")
@click.pass_obj
def cli_release_deploy(obj):
    Release("deploy", obj["project"], obj["config"], obj["tokens"]).run()


@release.command("rollback", help="Rollback a deployment")
@click.pass_obj
def cli_release_rollback(obj):
    Release("rollback", obj["project"], obj["config"], obj["tokens"]).run()


@main.command("deploy", help="Deploy master branch in production")
@click.pass_obj
def cli_deploy(obj):
    Deploy(obj["project"], obj["config"], obj["tokens"]).run()


@main.command("status", help="Current status of a project")
@click.pass_obj
def cli_status(obj):
    Status(obj["project"], obj["config"]).run()


@main.command("check", help="Verify authorisations for third party services")
@click.pass_obj
def cli_check(obj):
    Check(obj["config"], obj["tokens"]).run()


@main.command("id", help="Get the ID of the hosts where the task is running")
@click.argument("environment", type=click.Choice(("staging", "production")))
@click.pass_obj
def cli_id(obj, environment):
    ID(obj["project"], environment).run()


@main.command("ip", help="Get the IP addresses of the hosts where the task is running")
@click.argument("environment", type=click.Choice(("staging", "production")))
@click.pass_obj
def cli_ip(obj, environment):
    IP(obj["project"], environment).run()


@main.command("generator", help="Generate different files from templates")
@click.pass_obj
def cli_generator(obj):
    Generator(obj["project"], obj["config"], obj["tokens"]).run()


@main.command("aggregator", help="Manage CNAMEs of aggregators in QA envs")
@click.option("-l", "--list", "show_list", required=False, count=True)
@click.option("-c", "--change", "change", required=False, count=True)
@click.pass_obj
def cli_aggregator(obj, show_list, change):
    Aggregator(obj["config"], show_list, change).run()


@main.command("login", help="manage login against Auth0")
@click.pass_obj
def login(obj):
    Login(obj["config"]).run()


@main.command(
    "secret", help="Manage secrets grants in multiple countries (aws-vault needed)"
)
@click.option("-c", "--create", "create_action", required=False, count=True)
@click.option("-g", "--grant", "grant_action", required=False, count=True)
@click.option("-b", "--base-profile", "base_profile", required=False)
@click.pass_obj
def cli_secret(obj, create_action, grant_action, base_profile):
    Secret(
        obj["project"], obj["config"], create_action, grant_action, base_profile
    ).run()



# -*- encoding: utf-8 -*-
from termcolor import colored


def debug(message):
    print(colored(message, "blue"))


def info(message):
    print(colored(message, "green"))


def warning(message):
    print(colored(message, "yellow"))


def error(message):
    print(colored(message, "red"))



# -*- coding: utf-8 -*-
from halo import Halo

from suite_py.lib import logger
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.symbol import CHECKMARK, CROSSMARK


class Status:
    def __init__(self, project, config):
        self._project = project
        self._captainhook = CaptainHook(config)

    def run(self):
        with Halo(text="Contacting Captainhook...", spinner="dots", color="magenta"):
            staging_status = self._captainhook.status(self._project, "staging").json()
            production_status = self._captainhook.status(
                self._project, "production"
            ).json()

        _forge_message(staging_status, "staging")
        _forge_message(production_status, "production")


def _forge_message(status, env):
    if status["locked"]:
        if status["by"] == "":
            logger.error(f"{CROSSMARK} {env}\n      locked")
        else:
            logger.error(f"{CROSSMARK} {env}\n      locked by {status['by']}")
    else:
        logger.info(f"{CHECKMARK} {env}\n      not locked")



# -*- encoding: utf-8 -*-
import subprocess
import yaml

from suite_py.lib import logger
from suite_py.lib.handler.git_handler import GitHandler

REPO = "qainit"


def get_qa_projects(config):
    git = GitHandler(REPO, config)
    git.check_repo_cloned()
    git.sync()
    with open(f"{git.get_path()}/projects.yml", "r") as file:
        branches_obj = yaml.safe_load(file.read())

    return list(branches_obj.keys())


def qainit_deploy(args, config):
    qainit_dir = f"{config.user['projects_home']}/{REPO}"
    with open(f"{qainit_dir}/suitepy-projects.yml", "w+") as file:
        file.write(args)

    # development only
    # twig_command = f"{config.user['projects_home']}/twig-binaries/bin/twig-feature"
    return subprocess.run(
        # [twig_command, "suite", "deploy"], # development only
        ["twig", "feature", "suite", "deploy"],
        cwd=qainit_dir,
        check=True,
    )


def qainit_shutdown(youtrack_id, config):
    git = GitHandler(REPO, config)
    git.check_repo_cloned()
    git.sync()
    branch = git.search_remote_branch(f"*{youtrack_id}*")
    if branch:
        git.checkout(branch)
        git.commit("shutdown", dummy=True)
        git.push(branch)

    else:
        logger.warning(
            "I could not find a QA env for this card, if the QA exists, please shut it down manually.\nDevops are watching you! ( •͡˘ _•͡˘)"
        )
